#!/usr/bin/env python3

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from disdrodb.L0.L0_processing import run_L0


from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    column_names = [
        "id",
        "latitude",
        "longitude",
        "time",
        "datalogger_temperature",
        "datalogger_voltage",
        "rainfall_rate_32bit",
        "rainfall_accumulated_32bit",
        "weather_code_synop_4680",
        "weather_code_synop_4677",
        "reflectivity_32bit",
        "mor_visibility",
        "laser_amplitude",
        "number_particles",
        "sensor_temperature",
        "sensor_heating_current",
        "sensor_battery_voltage",
        "sensor_status",
        "rainfall_amount_absolute_32bit",
        "error_code",
        "raw_drop_concentration",
        "raw_drop_average_velocity",
        "raw_drop_number",
        "datalogger_error",
    ]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = ","

    # - Avoid first column to become df index
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error"]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"
    reader_kwargs["header"] = None

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards
    df_sanitizer_fun = None

    def df_sanitizer_fun(df, lazy=False):
        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        # - Drop datalogger columns
        columns_to_drop = [
            "id",
            "datalogger_temperature",
            "datalogger_voltage",
            "datalogger_error",
        ]

        df = df.drop(columns=columns_to_drop)

        # - Drop latitude and longitude
        # --> Latitude and longitude is specified in the the metadata.yaml
        df = df.drop(columns=["latitude", "longitude"])

        # - Convert time column to datetime with resolution in seconds
        df["time"] = dd.to_datetime(df["time"], format="%d-%m-%Y %H:%M:%S")

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files within raw_dir/data/<station_id>
    files_glob_pattern = "*.dat*"

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )



##################################################
## Wrapper to run readers by command lines
##################################################
## Author: Régis Longchamp
##################################################
import click

from disdrodb.L0 import click_L0_readers_options

# -------------------------------------------------------------------------.
# Click Command Line Interface decorator
@click.command()
@click.argument("data_source", metavar="<data_source>")
@click.argument("reader_name", metavar="<reader_name>")
@click_L0_readers_options  # get default arguments
def run_reader_cmd(**kwargs):
    """Wrapper to run reader functions by command lines.

    Parameters
    ----------
    data_source : str
        Institution name (when campaign data spans more than 1 country) or country (when all campaigns (or sensor networks) are inside a given country)
    reader_name : str
        Campaign name
    raw_dir : str
        The directory path where all the raw content of a specific campaign is stored.
        The path must have the following structure:
            <...>/DISDRODB/Raw/<data_source>/<campaign_name>'.
        Inside the raw_dir directory, it is required to adopt the following structure:
        - /data/<station_id>/<raw_files>
        - /metadata/<station_id>.yaml
        Important points:
        - For each <station_id> there must be a corresponding YAML file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files.
        - The campaign_name are expected to be UPPER CASE.
    processed_dir : str
        The desired directory path for the processed DISDRODB L0A and L0B products.
        The path should have the following structure:
            <...>/DISDRODB/Processed/<data_source>/<campaign_name>'
        For testing purpose, this function exceptionally accept also a directory path simply ending
        with <campaign_name> (i.e. /tmp/<campaign_name>).
    l0a_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.
    l0b_processing : bool
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data.
      The default is True.
    keep_l0a : bool
        Whether to keep the L0A files after having generated the L0B netCDF products.
        The default is False.
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        - For L0A processing, it processes just 3 raw data files.
        - For L0B processing, it takes a small subset of the L0A Apache Parquet dataframe.
        The default is False.
    lazy : bool
        Whether to perform processing lazily with dask.
        If lazy=True, it employed dask.array and dask.dataframe.
        If lazy=False, it employed pandas.DataFrame and numpy.array.
        The default is True.
    single_netcdf : bool
        Whether to concatenate all raw files into a single L0B netCDF file.
        If single_netcdf=True, all raw files will be saved into a single L0B netCDF file.
        If single_netcdf=False, each raw file will be converted into the corresponding L0B netCDF file.
        The default is True.
    """
    from disdrodb.L0.L0_processing import run_reader

    run_reader(**kwargs)


if __name__ == "__main__":
    run_reader_cmd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 17:42:56 2022

@author: ghiggi
"""
import os
import glob
import yaml
from typing import Union

# TODO: to improve a lot !!!
# - Check metadata !!!
# - Define rules !!!


def read_yaml(fpath: str) -> dict:
    """Read YAML file.

    Parameters
    ----------
    fpath : str
        Input YAML file path.

    Returns
    -------
    dict
        Attributes read from the YAML file.
    """
    with open(fpath, "r") as f:
        attrs = yaml.safe_load(f)
    return attrs


def identify_missing_metadata(metadata_fpaths: str, keys: Union[str, list]) -> None:
    """Identify missing metadata.

    Parameters
    ----------
    metadata_fpaths : str
        Input YAML file path.
    keys : Union[str,list]
        Attributes to verify the presence.
    """
    if isinstance(keys, str):
        keys = [keys]
    for key in keys:
        for fpath in metadata_fpaths:
            #  print(fpath)
            metadata = read_yaml(fpath)
            if len(str(metadata.get(key, ""))) == 0:  # ensure is string to avoid error
                print(f"Missing {key} at: ", fpath)
    return None


def identify_missing_coords(metadata_fpaths: str) -> None:
    """Identify missing coordinates.

    Parameters
    ----------
    metadata_fpaths : str
        Input YAML file path.

    Raises
    ------
    TypeError
        Error if latitude or longitude coordinates are not present or are wrongly formatted.

    """
    for fpath in metadata_fpaths:
        metadata = read_yaml(fpath)
        longitude = metadata.get("longitude", -9999)
        latitude = metadata.get("latitude", -9999)
        # Check type validity
        if isinstance(longitude, str):
            raise TypeError(f"longitude is not defined as numeric at {fpath}.")
        if isinstance(latitude, str):
            raise TypeError(f"latitude is not defined as numeric at {fpath}.")
        # Check is not none
        if isinstance(longitude, type(None)) or isinstance(latitude, type(None)):
            print(f"Unspecified lat lon coordinates at: {fpath}")
            continue
        # Check value validity
        if longitude == -9999 or latitude == -9999:
            print(f"Missing lat lon coordinates at: {fpath}")
        elif longitude > 180 or longitude < -180:
            print("Unvalid longitude at : ", fpath)
        elif latitude > 90 or latitude < -90:
            print("Unvalid latitude at : ", fpath)
        else:
            pass
    return None


# # EXAMPLE
# ARCHIVE_DIR = "/ltenas3/0_Data/DISDRODB/Raw" # ARCHIVE_DIR = "/ltenas3/0_Data/DISDRODB/Processed"
# metadata_fpaths = glob.glob(os.path.join(ARCHIVE_DIR, "*/*/metadata/*.yml"))
# identify_missing_coords(metadata_fpaths)

# identify_missing_metadata(metadata_fpaths, keys="station_id")
# identify_missing_metadata(metadata_fpaths, keys="station_name")
# identify_missing_metadata(metadata_fpaths, keys=["station_id","station_name"])
# identify_missing_metadata(metadata_fpaths, keys="campaign_name")

# -----------------------------------------------------------------------------.
# TODO
# - station_id
# - station_number --> to be replaced by station_id
# - station_name --> if missing, use station_id

# - TODO: define dtype of metadata !!!
# - raise error if missing station_id
# - raise error if missing campaign_name
#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 17:07:56 2022

@author: ghiggi
"""
import numpy as np
from disdrodb.L0.standards import (
    get_diameter_bin_center,
    get_diameter_bin_lower,
    get_diameter_bin_upper,
    get_diameter_bin_width,
    get_velocity_bin_center,
    get_velocity_bin_lower,
    get_velocity_bin_upper,
    get_velocity_bin_width,
    get_description_dict,
    get_units_dict,
    get_long_name_dict,
    get_data_format_dict,
    get_L0B_encodings_dict,
)

# ------------------------------------------------------------
# TODO:
# - check variables in L0A_dtypes match L0B_encodings.yml keys

# ------------------------------------------------------------

sensor_name = "OTT_Parsivel"  # LITTLE ISSUE WHEN IT STARTS THE DIAMETER ...
sensor_name = "OTT_Parsivel2"  # LITTLE ISSUE WHEN IT STARTS THE DIAMETER ...
sensor_name = "Thies_LPM"  # OK


def check_bin_consistency(sensor_name: str) -> None:
    """Check bin consistency from config file.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.
    """

    diameter_bin_lower = get_diameter_bin_lower(sensor_name)
    diameter_bin_upper = get_diameter_bin_upper(sensor_name)
    diameter_bin_center = get_diameter_bin_center(sensor_name)
    diameter_bin_width = get_diameter_bin_width(sensor_name)
    diameter_bin_lower = np.array(diameter_bin_lower)
    diameter_bin_upper = np.array(diameter_bin_upper)
    diameter_bin_center = np.array(diameter_bin_center)
    diameter_bin_width = np.array(diameter_bin_width)

    np.testing.assert_allclose(
        diameter_bin_upper - diameter_bin_lower, diameter_bin_width
    )
    np.testing.assert_allclose(
        diameter_bin_lower + diameter_bin_width / 2, diameter_bin_center
    )
    np.testing.assert_allclose(
        diameter_bin_upper - diameter_bin_width / 2, diameter_bin_center
    )

    velocity_bin_lower = get_velocity_bin_lower(sensor_name)
    velocity_bin_upper = get_velocity_bin_upper(sensor_name)
    velocity_bin_center = get_velocity_bin_center(sensor_name)
    velocity_bin_width = get_velocity_bin_width(sensor_name)

    velocity_bin_lower = np.array(velocity_bin_lower)
    velocity_bin_upper = np.array(velocity_bin_upper)
    velocity_bin_center = np.array(velocity_bin_center)
    velocity_bin_width = np.array(velocity_bin_width)

    np.testing.assert_allclose(
        velocity_bin_upper - velocity_bin_lower, velocity_bin_width
    )
    np.testing.assert_allclose(
        velocity_bin_lower + velocity_bin_width / 2, velocity_bin_center
    )
    np.testing.assert_allclose(
        velocity_bin_upper - velocity_bin_width / 2, velocity_bin_center
    )


def check_variable_keys_consistency(sensor_name: str) -> None:
    """Check attributes consistency from config file.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.
    """
    description_dict = get_description_dict(sensor_name)
    units_dict = get_units_dict(sensor_name)
    long_name_dict = get_long_name_dict(sensor_name)
    data_format_dict = get_data_format_dict(sensor_name)
    encoding_dict = get_L0B_encodings_dict(sensor_name)
    encoding_vars = set(encoding_dict.keys())
    data_format_vars = set(data_format_dict.keys())
    long_name_vars = set(long_name_dict.keys())
    units_vars = set(units_dict.keys())
    description_vars = set(description_dict.keys())

    encoding_vars.difference(data_format_vars)
    encoding_vars.difference(units_vars)
    encoding_vars.difference(description_vars)
    # encoding_vars.difference(long_name_vars) # TODO ADD

    data_format_vars.difference(encoding_vars)
    units_vars.difference(encoding_vars)
    description_vars.difference(encoding_vars)
    # long_name_vars.difference(encoding_vars) # TODO ADD


# get_available_sensor_name()

# get_diameter_bins_dict(sensor_name)
# get_velocity_bins_dict(sensor_name)

# get_variables_dict(sensor_name)
# get_sensor_variables(sensor_name)

# get_units_dict(sensor_name)
# get_explanations_dict(sensor_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 23 14:02:12 2022

@author: ghiggi
"""
import dask
import logging
import numpy as np
import xarray as xr

logger = logging.getLogger(__name__)


def ensure_monotonic_dimension(fpaths: str, list_ds: list, dim: str = "time") -> list:
    """Ensure that a list of xr.Dataset has a monotonic increasing dimension.

    Parameters
    ----------
    fpaths : list
        List of netCDFs file paths.
    list_ds : list
        List of xarray Dataset.
    dim : str, optional
        Dimension name.
        The default is "time".

    Returns
    -------
    list
        List of xarray datasets.
    """

    # Get dimension values array (and associated list_ds/xr.Dataset indices)
    dim_values = np.concatenate([ds[dim].values for ds in list_ds])
    list_index = np.concatenate(
        [np.ones(len(ds[dim])) * i for i, ds in enumerate(list_ds)]
    )
    list_index = list_index.astype(int)
    ds_index = np.concatenate(
        [np.arange(0, len(ds[dim])) for i, ds in enumerate(list_ds)]
    )

    # Identify index where start decreasing or duplicated
    diff_dim_values = np.diff(dim_values)
    indices_decreasing = np.where(diff_dim_values.astype(float) <= 0)[0] + 1

    if len(indices_decreasing) > 0:
        idx_start_decreasing = indices_decreasing[0]  # or duplicate

        # Find last timestep that is lower or equal to that timestep
        idx_restart_increase = np.max(
            np.where(dim_values <= dim_values[idx_start_decreasing - 1])[0]
        )

        # Indices to remove
        idx_to_remove = np.arange(idx_start_decreasing, idx_restart_increase + 1)
        list_index_bad = list_index[idx_to_remove]
        ds_index_bad = ds_index[idx_to_remove]
        dim_values_bad = dim_values[idx_to_remove]

        # Get dictionary of values/idx which are duplicated (or lead to no monotonic dimension)
        dict_ds_bad_values = {
            k: dim_values_bad[np.where(list_index_bad == k)[0]]
            for k in np.unique(list_index_bad)
        }
        dict_ds_bad_idx = {
            k: ds_index_bad[np.where(list_index_bad == k)[0]]
            for k in np.unique(list_index_bad)
        }

        # Print message
        for ds_idx_bad, bad_values in dict_ds_bad_values.items():
            fpath = fpaths[ds_idx_bad]
            msg = f"In {fpath}, dropping {dim} values {bad_values} to ensure monotonic {dim} dimension."
            logger.debug(msg)

        # Remove non-unique and not  da
        for ds_idx_bad, bad_idx in dict_ds_bad_idx.items():
            ds = list_ds[ds_idx_bad]
            list_ds[ds_idx_bad] = ds.drop_isel({dim: bad_idx})

        # Iterative check
        list_ds = ensure_monotonic_dimension(fpaths, list_ds, dim=dim)

    # Return list of xr.Dataset with monotonic dimension
    return list_ds


# ds_index = [0,1,2,3,0,1,2,3,4]
# list_index = [0,0,0,0,1, 1, 1,1, 1]
# dim_values = [0,1,5,5,5, 5, 6,7,8]
# list_index = np.array(list_index)
# dim_values = np.array(dim_values)
# ds_index = np.array(ds_index)


def get_common_coords(list_ds: list) -> set:
    """Get the common set of coordinates across xarray datasets.

    Parameters
    ----------
    list_ds : list
        List of xarray datasets.

    Returns
    -------
    set
        The common set of coordinates.
    """
    # Get common data vars
    coords_ref = set(list_ds[0].coords)
    for ds in list_ds:
        coords_ref = set(ds.coords).intersection(coords_ref)
    return coords_ref


def get_common_vars(list_ds: list) -> tuple:
    """Get the common set of variables across xarray datasets.

    Parameters
    ----------
    list_ds : list
        List of xarray datasets

    Returns
    -------
    tuple
        (The common set of variables, Dictionary to collect problems)
    """
    # Initialize variables in first file
    vars_init = set(list_ds[0].data_vars)
    n_vars_init = len(vars_init)
    # Initialize common variable reference
    common_vars_ref = set(list_ds[0].data_vars)
    # Initialize dictionary to collect problems
    dict_problems = {}
    dict_problems["missing_versus_first"] = {}
    dict_problems["additional_versus_first"] = {}
    dict_problems["evolution"] = {}
    dict_problems["missing_versus_first"]["ref"] = vars_init
    dict_problems["additional_versus_first"]["ref"] = vars_init
    dict_problems["evolution"]["ref"] = vars_init
    # Loop over all datasets
    for i, ds in enumerate(list_ds):
        # Extract current dataset variable info
        vars_ds = set(ds.data_vars)
        n_vars = len(vars_ds)
        # Check if missing variable compared to first
        vars_intersection_vs_first = vars_ds.intersection(vars_init)
        # Collect information on missing variables (compared to first)
        if len(vars_intersection_vs_first) != n_vars_init:
            if len(vars_init.difference(vars_ds)) > 0:
                dict_problems["missing_versus_first"][i] = vars_init.difference(vars_ds)
            if len(vars_ds.difference(vars_init)) > 0:
                dict_problems["additional_versus_first"][i] = vars_ds.difference(
                    vars_init
                )
        # Check if missing an additional variable compared to past common variables
        common_vars_ref_new = vars_ds.intersection(common_vars_ref)
        # If missing, collect information on additional missing variable
        if len(common_vars_ref_new) != len(common_vars_ref):
            dict_problems["evolution"][i] = common_vars_ref.difference(
                common_vars_ref_new
            )

        # Redefine variables common to all datasets
        common_vars_ref = common_vars_ref_new

    # ---------------------------------------------.
    # Return results
    return common_vars_ref, dict_problems


def get_list_ds(fpaths: str) -> list:
    """Get list of xarray datasets from file paths.

    Parameters
    ----------
    fpaths : list
        List of netCDFs file paths.

    Returns
    -------
    list
        List of xarray datasets.
    """

    @dask.delayed
    def open_dataset_delayed(fpath):
        ds = xr.open_dataset(fpath, chunks={"time": -1}, engine="netcdf4")
        return ds

    list_ds_delayed = []
    for fpath in fpaths:
        list_ds_delayed.append(open_dataset_delayed(fpath))
    list_ds = dask.compute(list_ds_delayed)[0]
    return list_ds


def ensure_constant_coords(list_ds: list, coords: list) -> tuple:
    """Check coordinates are invariant across xarray datasets.

    Parameters
    ----------
    list_ds : list
        List of xarray datasets.
    coords : list
        List of coordinates names.

    Returns
    -------
    tuple
        (List of xarray datasets, Dictionary to collect problems)
    """
    dict_problems = {}
    for coord in coords:
        ref_set = set(list_ds[0][coord].values)
        dict_coord = {}
        dict_coord["ref"] = ref_set
        for i, ds in enumerate(list_ds):
            values = set(ds[coord].values)
            if ref_set != values:
                dict_coord[i] = values
                list_ds[i] = ds.assign_coords({coord: list(ref_set)})
        dict_problems[coord] = dict_coord
    return list_ds, dict_problems


def xr_concat_datasets(fpaths: str) -> xr.Dataset:
    """Concat xr.Dataset in a robust and parallel way.

    1. It checks for time dimension monotonicity
    2. It checks by default for a minimum common set of variables
    Note: client = Client(processes=True) to execute it fast !


    Parameters
    ----------
    fpaths : list
        List of netCDFs file paths.

    Returns
    -------
    xr.Dataset
        A single xarray dataset.

    Raises
    ------
    ValueError
        Error if the merging/concatenation operations can not be achieved.

    """

    # --------------------------------------.
    # Open xr.Dataset lazily in parallel using dask delayed
    list_ds = get_list_ds(fpaths)

    # --------------------------------------.
    # Ensure time dimension is monotonic increasingly
    list_ds = ensure_monotonic_dimension(fpaths, list_ds, dim="time")

    # --------------------------------------.
    # Ensure coordinate outside time do not vary
    coords_var = get_common_coords(list_ds)
    coords_without_time = coords_var.difference(
        set(["time", "latitude", "longitude", "lat", "lon"])
    )
    list_ds, dict_problems = ensure_constant_coords(list_ds, coords=coords_without_time)
    # --------------------------------------.
    # - Log the possible problems
    for coord in coords_without_time:
        dict_problem = dict_problems[coord]
        ref = dict_problem.pop("ref")
        n_problems = len(dict_problem)
        n_files = len(list_ds)
        if len(dict_problem) != 0:
            msg = f"Difference found in {n_problems}/{n_files} files."
            logger.debug(msg)
            msg = f"In the first file, the coordinate {coord} has values {ref}."
            logger.debug(msg)
            for i, values in dict_problem.items():
                fpath = fpaths[i]
                msg = f"In {fpath}, the coordinate {coord} has values {values}."
                logger.debug(msg)
                logger.info(
                    "Such coordinates have been replaced with the one observed in the first file."
                )

    # --------------------------------------.
    # Get set of common_vars
    common_vars, dict_problems = get_common_vars(list_ds)

    # --------------------------------------.
    # Log the missing variables compared to first file
    dict_problem = dict_problems["missing_versus_first"]
    n_problems = len(dict_problem)
    n_files = len(list_ds)
    ref = dict_problem.pop("ref")
    if len(dict_problem) != 0:
        msg = f"Difference found in {n_problems}/{n_files} files."
        for i, vars in dict_problem.items():
            fpath = fpaths[i]
            msg = f"At file {i}/{n_files} ({fpath}), the variables {vars} are missing (compared to first file)."
            logger.debug(msg)

    # --------------------------------------.
    # Log the additional variables compared to first file
    dict_problem = dict_problems["additional_versus_first"]
    n_problems = len(dict_problem)
    n_files = len(list_ds)
    ref = dict_problem.pop("ref")
    if len(dict_problem) != 0:
        msg = f"Difference found in {n_problems}/{n_files} files."
        for i, vars in dict_problem.items():
            fpath = fpaths[i]
            msg = f"At file {i}/{n_files} ({fpath}), there are the variables {vars} which are missing in the first file."
            logger.debug(msg)

    # --------------------------------------.
    # Log the progressively shrinkage of common variables
    dict_problem = dict_problems["evolution"]
    n_problems = len(dict_problem)
    n_files = len(list_ds)
    ref = dict_problem.pop("ref")
    if len(dict_problem) != 0:
        msg = "Here we report the file which led to shrinkage of the set of common variables."
        logger.debug(msg)
        for i, removed_vars in dict_problem.items():
            fpath = fpaths[i]
            for var in removed_vars:
                msg = f"At file {i}/{n_files} ({fpath}), the {var} is removed from the common set of variables."
                logger.debug(msg)

    # --------------------------------------.
    # Concat/merge all netCDFs
    # - If there are common variables, use xr.concat
    if len(common_vars) > 0:
        # Ensure common set of variables across xr.Datasets
        for i, ds in enumerate(list_ds):
            list_ds[i] = ds[common_vars]
        # Try concatenating
        try:
            logger.info("Start concatenating with xr.concat.")
            ds = xr.concat(list_ds, dim="time", coords="minimal", compat="override")
            logger.info("Concatenation with xr.concat has been successful.")
        except:
            msg = "Concatenation with xr.concat failed."
            logger.error(msg)
            raise ValueError(msg)
    # - Otherwise use xr.merge
    else:
        try:
            logger.info("Start concatenating with xr.merge.")
            ds = xr.merge(
                list_ds, compat="override", join="outer", combine_attrs="override"
            )
            logger.info("Concatenation with xr.merge has been successful.")
        except:
            msg = "Concatenation with xr.merge failed."
            logger.error(msg)
            raise ValueError(msg)
    # --------------------------------------.
    # Return xr.Dataset
    return ds



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan  1 17:18:04 2022

@author: ghiggi
"""
import os
import subprocess
from disdrodb.pipeline.utils_cmd import get_reader_cmd
from pathlib import Path

# You need to set the disdrodb repo path in your .bashrc
# export PYTHONPATH="${PYTHONPATH}:/home/sguzzo/Projects/disdrodb"
# You need to activate the disdrodb environment: conda activate disdrodb

# -----------------------------------------------------------------------------.
#### Define campaign dictionary
DELFT_dict = {
    "CABAUW": "reader_RASPBERRY.py",
}

#### Define filepaths
home = str(Path.home())
reader_dir = "/home/sguzzo/PycharmProjects/disdrodb/disdrodb/readers/DELFT"
raw_base_dir = "/home/sguzzo/Parsivel/RAW_TELEGRAM"
processed_base_dir = "/tmp/Processed"
# processed_base_dir = "/tmp/Processed/DELFT"

#### Processing settings
l0a_processing = True
l0b_processing = True
force = True
verbose = True
debugging_mode = False  # da cambiare se vuoi processare piu dei primi 5 files
lazy = False
# write_zarr = False
keep_l0a = True
single_netcdf = True

#### Process all campaigns
for campaign_name in DELFT_dict.keys():
    reader_filepath = os.path.join(reader_dir, DELFT_dict[campaign_name])
    cmd = get_reader_cmd(
        reader_filepath=reader_filepath,
        raw_dir=os.path.join(raw_base_dir, campaign_name),
        processed_dir=os.path.join(processed_base_dir, campaign_name),
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
    )

    subprocess.run(cmd, shell=True)
    # os.system(cmd)

# -----------------------------------------------------------------------------.
# TODO:
# --> Useful to test changes to code do not crash other reader
# --> debugging_mode=True to speed up tests ;)

# -----------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import logging
import numpy as np
import pandas as pd
import dask.dataframe as dd
from typing import Union
from disdrodb.L0.standards import get_data_format_dict, get_L0A_dtype


logger = logging.getLogger(__name__)


def check_sensor_name(sensor_name: str) -> None:
    """Check sensor name.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Raises
    ------
    TypeError
        Error if `sensor_name` is not a string.
    ValueError
        Error if the input sensor name has not been found in the list of available sensors.
    """
    from disdrodb.L0.standards import get_available_sensor_name

    available_sensor_name = get_available_sensor_name()
    if not isinstance(sensor_name, str):
        logger.exception("'sensor_name' must be a string'")
        raise TypeError("'sensor_name' must be a string'")
    if sensor_name not in available_sensor_name:
        msg = f"Valid sensor_name are {available_sensor_name}"
        logger.exception(msg)
        raise ValueError(msg)


def check_L0A_column_names(
    df: Union[pd.DataFrame, dd.DataFrame], sensor_name: str
) -> None:
    """Checks that the dataframe columns respects DISDRODB standards.

    Parameters
    ----------
    df : Union[pd.DataFrame,dd.DataFrame]
        Input dataframe.
    sensor_name : str
        Name of the sensor.

    Raises
    ------
    ValueError
        Error if some columns do not meet the DISDRODB standards or if the 'time' column is missing in the dataframe.

    """

    # Get valid columns
    dtype_dict = get_L0A_dtype(sensor_name)
    valid_columns = list(dtype_dict)
    valid_columns = valid_columns + ["time"]
    valid_columns = set(valid_columns)
    # Get dataframe column names
    df_columns = list(df.columns)
    df_columns = set(df_columns)
    # --------------------------------------------
    # Check there aren't valid columns
    unvalid_columns = list(df_columns.difference(valid_columns))
    if len(unvalid_columns) > 0:
        msg = (
            f"The following columns do no met the DISDRODB standards: {unvalid_columns}"
        )
        logger.error(msg)
        raise ValueError(msg)
    # --------------------------------------------
    # Check time column is present
    if "time" not in df_columns:
        msg = "The 'time' column is missing in the dataframe."
        logger.error(msg)
        raise ValueError(msg)
    # --------------------------------------------
    return None


def check_L0A_standards(
    fpath: str, sensor_name: str, raise_errors: bool = False, verbose: bool = True
) -> None:
    """Checks that a file respects the DISDRODB L0A standards.

    Parameters
    ----------
    fpath : str
        Input Apache Parquet file path.
    sensor_name : str
        Name of the sensor.
    raise_errors : bool, optional
        If True: it raises an error in case the file does not respect DISDROB L0A standards.
        If False: it logs the errors but does not interrupt the checks.
        By default is set to False.
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is True.

    Raises
    ------
    ValueError
        Error if some columns have inconsistent values.

    """
    # Read parquet
    df = pd.read_parquet(fpath)
    # -------------------------------------
    # Check data range
    dict_field_value_range = get_field_value_range_dict(sensor_name)
    list_wrong_columns = []
    for column in df.columns:
        if column in list(dict_field_value_range.keys()):
            if dict_field_value_range[column] is not None:
                if not df[column].between(*dict_field_value_range[column]).all():
                    list_wrong_columns.append(column)
                    if raise_errors:
                        raise ValueError(
                            f"'column' {column} has values outside the expected data range."
                        )

    if verbose:
        if len(list_wrong_columns) > 0:
            print(
                " - This columns have values outside the expected data range:",
                list_wrong_columns,
            )
    # -------------------------------------
    # Check categorical data values
    dict_field_values = get_field_value_options_dict(sensor_name)
    list_wrong_columns = []
    list_msg = []
    for column in df.columns:
        if column in list(dict_field_values.keys()):
            if not df[column].isin(dict_field_values[column]).all():
                list_wrong_columns.append(column)
                if raise_errors:
                    msg = f"'column' {column} has values different from {dict_field_values[column]}"
                    list_msg.append(msg)
                    raise ValueError(msg)
    if verbose:
        if len(list_wrong_columns) > 0:
            print(
                " - The following columns have values outside the expected data range:",
                list_wrong_columns,
            )
            [print(msg) for msg in list_msg]
    # -------------------------------------
    # Check if latitude and longitude are columns of the dataframe
    # - They should be only provided if the instrument is moving !!!!
    if "latitude" in df.columns:
        msg = " - The L0A dataframe has column 'latitude'. "
        "This should be included only if the sensor is moving. "
        "Otherwise, specify the 'latitude' in the metadata !"
        print(msg)
        logger.info(msg)

    if "longitude" in df.columns:
        msg = " - The L0A dataframe has column 'longitude'. "
        "This should be included only if the sensor is moving. "
        "Otherwise, specify the 'longitude' in the metadata !"
        print(msg)
        logger.info(msg)

    # -------------------------------------
    # Check if raw spectrum and 1D derivate exists
    list_sprectrum_vars = [
        "raw_drop_concentration",
        "raw_drop_average_velocity",
        "raw_drop_number",
    ]
    unavailable_vars = np.array(list_sprectrum_vars)[
        np.isin(list_sprectrum_vars, df.columns, invert=True)
    ]
    # Also if Thies_LPM has list_sprectrum_vars?
    if len(unavailable_vars) > 0:
        msg = (
            f" - The variables {unavailable_vars} are not present in the L0 dataframe."
        )
        print(msg)
        logger.info(msg)

    # -------------------------------------
    # Add index to dataframe
    # TODO

    # -------------------------------------
    # TODO[GG]:
    # if not respect standards, print errors and remove file (before raising errors)
    # - log, verbose ... L0A conforms to DISDRODB standards ;)

    # -------------------------------------


def check_L0B_standards(x: str) -> None:
    # TODO:
    # - Check for realistic values after having removed the flags !!!!
    pass


####--------------------------------------------------------------------------.
#### Get instrument default string standards
def get_field_ndigits_natural_dict(sensor_name: str) -> dict:
    """Get number of digits on the left side of the comma from the instrument default string standards.

    Example: 123,45 -> 123 --> 3 natural digits

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary with the expected number of natural digits for each data field.
    """

    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_naturals"] for k, v in data_dict.items()}
    return d


def get_field_ndigits_decimals_dict(sensor_name: dict) -> dict:
    """Get number of digits on the right side of the comma from the instrument default string standards.

    Example: 123,45 -> 45 --> 2 decimal digits
    Parameters
    ----------
    sensor_name : dict
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary with the expected number of decimal digits for each data field.
    """

    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_decimals"] for k, v in data_dict.items()}
    return d


def get_field_ndigits_dict(sensor_name: str) -> dict:
    """Get number of digits from the instrument default string standards.

    Important note: it excludes the comma but it counts the minus sign !!!


    Parameters
    ----------
    sensor_name : str
        Name of the sensor.
    Returns
    -------
    dict
        Dictionary with the expected number of digits for each data field.
    """

    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_digits"] for k, v in data_dict.items()}
    return d


def get_field_nchar_dict(sensor_name: str) -> dict:
    """Get the total number of characters from the instrument default string standards.

    Important note: it accounts also for the comma and the minus sign !!!


    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary with the expected number of characters for each data field.
    """

    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_characters"] for k, v in data_dict.items()}
    return d


def get_field_value_range_dict(sensor_name: str) -> dict:
    """Get the variable data range.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary with the expected data value range for each data field.
    """

    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["data_range"] for k, v in data_dict.items()}
    return d


def get_field_flag_dict(sensor_name: str) -> dict:
    """Get the variable nan flags.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary with the expected nan flag for each data field.
    """

    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["nan_flags"] for k, v in data_dict.items()}
    return d


# TODO: get_field_value_realistic_range  # when removing flags

####--------------------------------------------------------------------------.
#### Instrument default numeric standards
# TODO: this is currently used?
# --> YAML file with a list of variables providing error infos?
# --> And provide values that represents errors?


def get_field_value_options_dict(sensor_name: str) -> dict:
    """Get the dictionary of field values.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary of field values.

    Raises
    ------
    NotImplementedError
        Error if the name of the sensor is not available.
    """
    if sensor_name == "OTT_Parsivel":
        value_dict = {
            "sensor_status": [0, 1, 2, 3],
            "error_code": [0, 1, 2],
            # TODO: weather codes
            # Faculative/custom fields
            # 'datalogger_temperature': ['NaN']
            "datalogger_voltage": ["OK", "KO"],
            "datalogger_error": [0],
        }
    elif sensor_name == "OTT_Parsivel2":
        value_dict = {
            "sensor_status": [0, 1, 2, 3],
            "error_code": [0, 1, 2],
            # TODO: weather codes
            # Faculative/custom fields
        }
    elif sensor_name == "Thies_LPM":
        value_dict = {
            "laser_status": [0, 1],
            "laser_temperature_analog_status": [0, 1],
            "laser_temperature_digital_status": [0, 1],
            "laser_current_analog_status": [0, 1],
            "laser_current_digital_status": [0, 1],
            "sensor_voltage_supply_status": [0, 1],
            "current_heating_pane_transmitter_head_status": [0, 1],
            "current_heating_pane_receiver_head_status": [0, 1],
            "temperature_sensor_status": [0, 1],
            "current_heating_voltage_supply_status": [0, 1],
            "current_heating_house_status": [0, 1],
            "current_heating_heads_status": [0, 1],
            "current_heating_carriers_status": [0, 1],
            "control_output_laser_power_status": [0, 1],
            "reserve_status": [0, 1]
            # TODO: weather codes
            # Faculative/custom fields
        }
    else:
        raise NotImplementedError

    return value_dict


def get_field_error_dict(device: str) -> dict:
    """Get field error dictionnary

    Parameters
    ----------
    device : str
        Name of the sensor

    Returns
    -------
    dict
        Dictionnary of the field error
    """
    if device == "OTT_Parsivel":
        flag_dict = {
            "sensor_status": [1, 2, 3],
            "datalogger_error": [1],
            "error_code": [1, 2],
        }
    return flag_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
from disdrodb.L0 import run_L0


from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    # Notes
    # - In all files, the datalogger voltage hasn't the delimeter,
    #   so need to be split to obtain datalogger_voltage and rainfall_rate_32bit
    column_names = ["TO_SPLIT"]
    ##----------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}

    # - Define delimiter
    reader_kwargs["delimiter"] = "\\n"

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # Skip first row as columns names
    reader_kwargs["header"] = None

    # Skip the first row (header)
    reader_kwargs["skiprows"] = 1

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # Define encoding
    reader_kwargs["encoding"] = "latin1"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error"]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    def df_sanitizer_fun(df, lazy=False):
        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        # The delimiter ; is used for separating both the variables and the
        #   values of the raw spectrum.  So we need to retrieve the columns
        #   inside the sanitizer assuming a fixed number of columns.
        df = df["TO_SPLIT"].str.split(";", expand=True, n=16)

        # Define the column names
        column_names = [
            "date",
            "time",
            "rainfall_rate_32bit",
            "rainfall_accumulated_32bit",
            "reflectivity_32bit",
            "mor_visibility",
            "laser_amplitude",
            "number_particles",
            "sensor_temperature",
            "sensor_heating_current",
            "sensor_battery_voltage",
            "rain_kinetic_energy",
            "snowfall_rate",
            "weather_code_synop_4680",
            "weather_code_metar_4678",
            "weather_code_nws",
            "raw_drop_number",
        ]
        df.columns = column_names

        # Define the time column
        df["time"] = df["date"] + "-" + df["time"]
        df["time"] = dd.to_datetime(df["time"], format="%Y/%m/%d-%H:%M:%S")
        df = df.drop(columns=["date"])

        # Preprocess the raw spectrum
        # - The '<SPECTRUM>ZERO</SPECTRUM>'  indicates no drops detected
        # - So replace the string with '' so that L0B processing generate a matrix filled by 0s.
        df["raw_drop_number"] = df["raw_drop_number"].str.replace(
            "<SPECTRUM>ZERO</SPECTRUM>", "''"
        )
        # Remove <SPECTRUM> and </SPECTRUM>" acroynms from the raw_drop_number field
        df["raw_drop_number"] = df["raw_drop_number"].str.replace("<SPECTRUM>", "")
        df["raw_drop_number"] = df["raw_drop_number"].str.replace("</SPECTRUM>", "")

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    files_glob_pattern = "*.txt"  # There is only one file without extension

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=False,  # The actual solution work only with pandas, to change in the future
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )


if __name__ == "__main__":
    reader()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 13:38:37 2022

@author: ghiggi
"""
import os
import subprocess
from disdrodb.pipeline.utils_cmd import get_reader_cmd


# You need to set the disdrodb repo path in your .bashrc
# export PYTHONPATH="${PYTHONPATH}:/home/ghiggi/Projects/disdrodb"
# You need to activate the disdrodb envirnment: conda activate disdrodb

# -----------------------------------------------------------------------------.
#### Define campaign dictionary
# ARM_ld --> OTT Parsivel 2
# ARM_lpm --> Thies LPM
campaign_dict = {
    # "ALASKA": "reader_ARM_lpm.py",
    "ACE_ENA": "reader_ARM_ld.py",
    "AWARE": "reader_ARM_ld.py",
    "CACTI": "reader_ARM_ld.py",
    "COMBLE": "reader_ARM_ld.py",
    "GOAMAZON": "reader_ARM_ld.py",
    "MARCUS": "reader_ARM_ld.py",  # MARCUS S1, MARCUS S2 are mobile ...
    "MICRE": "reader_ARM_ld.py",
    "MOSAIC": "reader_ARM_ld.py",  # MOSAIC M1, MOSAIC S3 are mobile ...
    "SAIL": "reader_ARM_ld.py",
    "SGP": "reader_ARM_ld.py",
    "TRACER": "reader_ARM_ld.py",
}

#### Define filepaths
reader_dir = "/ltenas3/0_Projects/disdrodb/disdrodb/readers/ARM"  # TO CHANGE
raw_base_dir = "/ltenas3/0_Data/DISDRODB/Raw/ARM"
processed_base_dir = "/ltenas3/0_Data/DISDRODB/Processed/ARM"
# processed_base_dir = "/tmp/DISDRODB/ARM"

#### Processing settings
force = True
verbose = True
debugging_mode = False
lazy = True

#### Process all campaigns
campaign_name = list(campaign_dict.keys())[0]
for campaign_name in campaign_dict.keys():
    print("Processing: ", campaign_name)
    reader_filepath = os.path.join(reader_dir, campaign_dict[campaign_name])
    raw_dir = os.path.join(raw_base_dir, campaign_name)
    processed_dir = os.path.join(processed_base_dir, campaign_name)
    cmd = get_reader_cmd(
        reader_filepath=reader_filepath,
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
    )

    print(subprocess.run(cmd, shell=True))
    os.system(cmd)

# -----------------------------------------------------------------------------.
# TODO:
# --> Useful to test changes to code do not crash other reader
# --> debuggin_mode=True to speed up tests ;)

# -----------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  4 10:53:10 2022

@author: kimbo
"""
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
from disdrodb.L0 import run_L0


def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):
    """Script to process raw data to L0A and L0B format.

    Parameters
    ----------
    raw_dir : str
        Directory path of raw file for a specific campaign.
        The path should end with <campaign_name>.
        Example raw_dir: '<...>/disdrodb/data/raw/<campaign_name>'.
        The directory must have the following structure:
        - /data/<station_id>/<raw_files>
        - /metadata/<station_id>.json
        Important points:
        - For each <station_id> there must be a corresponding JSON file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files.
        - The campaign_name are set to be UPPER CASE.
    processed_dir : str
        Desired directory path for the processed L0A and L0B products.
        The path should end with <campaign_name> and match the end of raw_dir.
        Example: '<...>/disdrodb/data/processed/<campaign_name>'.
    L0A_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.
    L0B_processing : bool
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data.
      The default is True.
    keep_L0A : bool
        Whether to keep the L0A files after having generated the L0B netCDF products.
        The default is False.
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        - For L0A processing, it processes just 3 raw data files.
        - For L0B processing, it takes a small subset of the L0A Apache Parquet dataframe.
        The default is False.
    lazy : bool
        Whether to perform processing lazily with dask.
        If lazy=True, it employed dask.array and dask.dataframe.
        If lazy=False, it employed pandas.DataFrame and numpy.array.
        The default is True.
    single_netcdf : bool
        Whether to concatenate all raw files into a single L0B netCDF file.
        If single_netcdf=True, all raw files will be saved into a single L0B netCDF file.
        If single_netcdf=False, each raw file will be converted into the corresponding L0B netCDF file.
        The default is True.

    """
    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    # Notes
    # - In all files, the datalogger voltage hasn't the delimeter,
    #   so need to be split to obtain datalogger_voltage and rainfall_rate_32bit

    # Columns defined into sanitizer
    column_names = []

    ##------------------------------------------------------------------------.
    #### - Define reader options

    reader_kwargs = {}

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = [
        "na",
        "",
        "error",
        "NA",
        "-.-",
        "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
    ]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    # Cast all to string
    reader_kwargs["dtype"] = str

    # No header
    reader_kwargs["header"] = None

    # Different enconding for this campaign
    reader_kwargs["encoding"] = "latin-1"  # Important for this campaign

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    df_sanitizer_fun = None

    def df_sanitizer_fun(df, lazy=lazy):

        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        temp_time = df.loc[df.iloc[:, 0].astype(str).str.len() == 16].add_prefix("col_")
        temp_time["col_0"] = dd.to_datetime(temp_time["col_0"], format="%Y.%m.%d;%H:%M")

        # Insert Raw into a series and drop last line
        temp_raw = df.loc[df.iloc[:, 0].astype(str).str.len() != 16].add_prefix("col_")
        temp_raw["col_0"] = temp_raw["col_0"].str.lstrip("   ")
        temp_raw = temp_raw.dropna()

        # If raw_drop_number series is not a 32 multiple, throw error
        if len(temp_raw) % 32 != 0:
            msg = "Wrong column number on raw_drop_number, can not parse!"
            raise ValueError(msg)

        # Series and variable temporary for parsing raw_drop_number
        if lazy:
            import pandas as pd

            raw = pd.DataFrame({"raw_drop_number": []})
            # raw = dd.from_pandas(raw, npartitions=1, chunksize=None)
        else:
            raw = pd.DataFrame({"raw_drop_number": []})
        temp_string_2 = ""

        # Parse for raw_drop_number
        for index, value in temp_raw.iterrows():
            temp_string = ""

            if index % 32 != 0:

                temp_string = value["col_0"].split(" ")

                # Remove blank string from split
                temp_string = list(filter(None, temp_string))

                # Cast list to int
                temp_string = [int(i) for i in temp_string]

                # Join list separate by comma
                temp_string = ", ".join(map(str, temp_string))

                # Add last comma
                temp_string += ", "

                temp_string_2 += temp_string

            else:

                raw = raw.append({"raw_drop_number": temp_string_2}, ignore_index=True)

                temp_string_2 = ""

        if lazy:
            raw = dd.from_pandas(raw, npartitions=1, chunksize=None)

        # Reset all index
        temp_time = temp_time.reset_index(drop=True)
        raw = raw.reset_index(drop=True)

        if lazy:
            df = dd.concat([temp_time, raw], axis=1, ignore_unknown_divisions=True)
        else:
            df = dd.concat([temp_time, raw], axis=1)

        df.columns = ["time", "raw_drop_number"]

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    files_glob_pattern = "*.txt*"

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from asyncio.log import logger
import os
import time
import logging


def create_L0_logger(
    processed_dir: str, campaign_name: str, verbose: bool = False
) -> logger:
    """Create L0 logger.

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory.
    campaign_name : str
        Campaign name.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.

    Returns
    -------
    logger
        Logger object.
    """
    # Define log name
    logger_name = "LO_" + "reader_" + campaign_name
    # Create logs directory
    logs_dir = os.path.join(processed_dir, "logs")
    os.makedirs(logs_dir, exist_ok=True)
    # Create logger
    _create_logger(logs_dir, logger_name)
    # Retrieve logger
    logger = logging.getLogger(campaign_name)
    # Update logger
    msg = "### Script started ###"
    if verbose:
        print("\n  " + msg + "\n")
    logger.info(msg)
    # Return logger
    return logger


def _create_logger(log_dir, logger_name):
    # Define log file filepath
    logger_fname = f'{time.strftime("%d-%m-%Y_%H-%M-%S")}_{logger_name}.log'
    logger_fpath = os.path.join(log_dir, logger_fname)
    # -------------------------------------------------------------------------.
    # Define logger format
    format_type = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    # Define logger level
    level = logging.DEBUG

    # Define logging
    logging.basicConfig(format=format_type, level=level, filename=logger_fpath)
    return None


def close_logger(logger: logger) -> None:
    """Close the logger

    Parameters
    ----------
    logger : logger
        Logger object.
    """
    handlers = logger.handlers[:]
    for handler in handlers:
        handler.close()
        logger.removeHandler(handler)
    return


# ----------------------------
def log_debug(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include debug entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.debug(msg)
    if verbose:
        print(" - " + msg)


def log_info(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include info entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.info(msg)
    if verbose:
        print(" - " + msg)


def log_warning(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include warning entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.warning(msg)
    if verbose:
        print(" - " + msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 14 11:01:52 2022

@author: kimbo
"""
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from disdrodb.L0 import run_L0

# -------------------------------------------------------------------------.
from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    # Notes
    # - In all files, the datalogger voltage hasn't the delimiter,
    #   so need to be split to obtain datalogger_voltage and rainfall_rate_32bit

    # "01","Rain intensity 32 bit",8,"mm/h","single_number"
    # "02","Rain amount accumulated 32 bit",7,"mm","single_number"
    # "03","Weather code SYNOP Table 4680",2,"","single_number"
    # "04","Weather code SYNOP Table 4677",2,"","single_number"
    # "05","Weather code METAR Table 4678",5,"","character_string"
    # "06","Weather code NWS",4,"","character_string"
    # "07","Radar reflectivity 32 bit",6,"dBZ","single_number"
    # "08","MOR visibility in precipitation",5,"m","single_number"
    # "09","Sample interval",5,"s","single_number"
    # "10","Signal amplitude of laser",5,"","single_number"
    # "11","Number of particles detected and validated",5,"","single_number"
    # "12","Temperature in sensor housing",3,"degree_Celsius","single_number"
    # "13","Sensor serial number",6,"","character_string"
    # "14","Firmware IOP",6,"","character_string"
    # "15","Firmware DSP",6,"","character_string"
    # "16","Sensor head heating current",4,"A","single_number"
    # "17","Power supply voltage",4,"V","single_number"
    # "18","Sensor status",1,"","single_number"
    # "19","Date/time measuring start",19,"DD.MM.YYYY_hh:mm:ss","character_string"
    # "20","Sensor time",8,"hh:mm:ss","character_string"
    # "21","Sensor date",10,"DD.MM.YYYY","character_string"
    # "22","Station name",4,"","character_string"
    # "23","Station number",4,"","character_string"
    # "24","Rain amount absolute 32 bit",7,"mm","single_number"
    # "25","Error code",3,"","character_string"
    # "26","Temperature PCB",3,"degree_Celsius","single_number"
    # "27","Temperature in right sensor head",3,"degree_Celsius","single_number"
    # "28","Temperature in left sensor head",3,"degree_Celsius","single_number"
    # "30","Rain intensity 16 bit max 30 mm/h",6,"mm/h","single_number"
    # "31","Rain intensity 16 bit max 1200 mm/h",6,"mm/h","single_number"
    # "32","Rain amount accumulated 16 bit",7,"mm","single_number"
    # "33","Radar reflectivity 16 bit",5,"dBZ","single_number"
    # "34","Kinetic energy",7,"J/(m2*h)","single_number"
    # "35","Snowfall intensity",7,"mm/h","single_number"
    # "60","Number of all particles detected",8,"","single_number"
    # "61","List of all particles detected",13,"","list"
    # "90","raw_drop_concentration",224,"","vector"
    # "91","raw_drop_average_velocity",224,"","vector"
    # "93","Raw data",4096,"","matrix"

    column_names = ["time", "epoch_time", "TO_BE_PARSED"]

    ##------------------------------------------------------------------------.
    #### - Define reader options

    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = ";"

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define parser engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"
    # reader_kwargs['zipped'] = False
    # reader_kwargs['zipped'] = True
    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = [
        "na",
        "",
        "error",
        "NA",
        "-.-",
        " NA",
    ]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    # Cast all to string
    reader_kwargs["dtype"] = str
    # reader_kwargs["file_name_to_read_zipped"] = 'tar'

    # Skip first row as columns names
    reader_kwargs["header"] = None

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    df_sanitizer_fun = None

    def df_sanitizer_fun(df, lazy=False):
        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        # Station 8 has all raw_drop_number corrupted, so it can't be use
        # Bug on rows 105000 and 110000 for station 7 (000NETDL07 and PAR007 device name) on dask
        if lazy:
            df = df.compute()

        # if (df['TO_BE_PARSED'].str.contains('000NETDL07')).any() | (df['TO_BE_PARSED'].str.contains('PAR007')).any():
        #     # df = df.loc[105000:110000]
        #     df.drop(df.index[105000:110000], axis=0, inplace=True)

        # # Remove rows that have a corrupted "TO_BE_PARSED" column
        # df = df.loc[df["TO_BE_PARSED"].astype(str).str.len() > 50]

        # Check date lenght
        df_date_with_error = df.loc[df["time"].astype(str).str.len() != 15]
        if len(df_date_with_error.index) > 0:
            raise Exception("Some dates do not have 15 characters")

        df = df.loc[df["time"].astype(str).str.len() == 15]

        df["time"] = dd.to_datetime(df["time"], format="%Y%m%d-%H%M%S")

        # Remove dupliacte time
        df = df.drop_duplicates(subset=["time"])

        # Extract the last column that contains the 37 remaining attributes
        df_concatenated_column = df["TO_BE_PARSED"].str.split(";", expand=True, n=99)

        # try:
        #     df['time'] = d
        # d.to_datetime(df['time'], format='%Y%m%d-%H%M%S')
        # except ValueError:
        #     df['time'] = dd.to_datetime(df['time'], format='%Y%m%d-%H%M%S', errors='coerce')
        #     df = df.loc[df.time.notnull()]

        # remove TO_BE_PARSED column
        df = df.drop(["TO_BE_PARSED"], axis=1)

        # Add names to columns
        column_names_concatenated_column = [
            "rainfall_rate_32bit",
            "rainfall_accumulated_32bit",
            "weather_code_synop_4680",
            "weather_code_synop_4677",
            "weather_code_metar_4678",
            "weather_code_nws",
            "reflectivity_32bit",
            "mor_visibility",
            "sample_interval",
            "laser_amplitude",
            "number_particles",
            "sensor_temperature",
            "sensor_serial_number",
            "firmware_iop",
            "firmware_dsp",
            "sensor_heating_current",
            "sensor_battery_voltage",
            "sensor_status",
            "date_time_measurement_start",
            "sensor_time",
            "sensor_date",
            "station_name",
            "station_number",
            "rainfall_amount_absolute_32bit",
            "error_code",
            "sensor_temperature_pcb",
            "sensor_temperature_receiver",
            "sensor_temperature_trasmitter",
            "rainfall_rate_16_bit_30",
            "rainfall_rate_16_bit_1200",
            "rainfall_accumulated_16bit",
            "reflectivity_16bit",
            "rain_kinetic_energy",
            "snowfall_rate",
            "number_particles_all",
            "number_particles_all_detected",
        ]

        # Remove unnecessary columns
        df_unconcatenated_column = df_concatenated_column.iloc[
            :, : len(column_names_concatenated_column)
        ]

        # Renamed columns
        df_unconcatenated_column = df_unconcatenated_column.set_axis(
            column_names_concatenated_column, axis=1, inplace=False
        )

        # Remove char from rain intensity
        df_unconcatenated_column["rainfall_rate_32bit"] = df_unconcatenated_column[
            "rainfall_rate_32bit"
        ].str.lstrip("b'")

        # Remove spaces on weather_code_metar_4678 and weather_code_nws
        df_unconcatenated_column["weather_code_metar_4678"] = df_unconcatenated_column[
            "weather_code_metar_4678"
        ].str.strip()
        df_unconcatenated_column["weather_code_nws"] = df_unconcatenated_column[
            "weather_code_nws"
        ].str.strip()

        # Add the comma on the raw_drop_concentration, raw_drop_average_velocity and raw_drop_number
        df_raw_drop_concentration = (
            df_concatenated_column.iloc[
                :, len(column_names_concatenated_column) - 1 : 67
            ]
            .apply(lambda x: ",".join(x.dropna().astype(str)), axis=1)
            .to_frame("raw_drop_concentration")
        )
        df_raw_drop_average_velocity = (
            df_concatenated_column.iloc[:, 67:-1]
            .apply(lambda x: ",".join(x.dropna().astype(str)), axis=1)
            .to_frame("raw_drop_average_velocity")
        )

        # Station 8 has \r\n' at end
        if (df_unconcatenated_column["station_name"] == "PAR008").any():
            df_raw_drop_number = (
                df_concatenated_column.iloc[:, -1:]
                .squeeze()
                .str.replace(r"(\w{3})", r"\1,", regex=True)
                .str.rstrip("\\r\\n'")
                .to_frame("raw_drop_number")
            )
        else:
            df_raw_drop_number = (
                df_concatenated_column.iloc[:, -1:]
                .squeeze()
                .str.replace(r"(\w{3})", r"\1,", regex=True)
                .str.rstrip("'")
                .to_frame("raw_drop_number")
            )

        # Concat all togheter
        df = dd.concat(
            [
                df,
                df_unconcatenated_column,
                df_raw_drop_concentration,
                df_raw_drop_average_velocity,
                df_raw_drop_number,
            ],
            axis=1,
        )

        # Drop invalid rows
        df = df.loc[df["raw_drop_concentration"].astype(str).str.len() == 223]
        df = df.loc[df["raw_drop_average_velocity"].astype(str).str.len() == 223]
        df = df.loc[df["raw_drop_number"].astype(str).str.len() == 4096]

        # Drop variables not required in L0 Apache Parquet
        todrop = [
            "firmware_iop",
            "firmware_dsp",
            "date_time_measurement_start",
            "sensor_time",
            "sensor_date",
            "station_name",
            "station_number",
            "sensor_serial_number",
            "epoch_time",
            "sample_interval",
            "sensor_serial_number",
            "number_particles_all_detected",
        ]

        df = df.drop(todrop, axis=1)

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    # files_glob_pattern= "*.tar.xz"
    files_glob_pattern = "*.csv"

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
        # Custom arguments of the parser
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:33:39 2022

@author: ghiggi
"""
import numpy as np
import pandas as pd
import xarray as xr
from xarray.core import dtypes


def regularize_dataset(
    ds: xr.Dataset, range_freq: str, tolerance=None, method=None, fill_value=dtypes.NA
):
    """Regularize a dataset across time dimension with uniform resolution."""
    start = ds.time.values[0]
    end = ds.time.values[-1]
    # start_date = pd.to_datetime(start).date() # to start at 00
    # end_date = pd.to_datetime(end).date() + datetime.timedelta(hours=23, minutes=57, seconds=30)
    new_time_index = pd.date_range(
        start=pd.to_datetime(start), end=pd.to_datetime(end), freq=range_freq
    )

    # Regularize dataset and fill with NA values
    ds_reindexed = ds.reindex(
        {"time": new_time_index},
        method=method,  # do not fill gaps
        tolerance=tolerance,  # mismatch in seconds
        fill_value=fill_value,
    )
    return ds_reindexed



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.


def get_DELFT_dict(sensor_name: str) -> dict:
    """
    Dictionary mapping from DELFT standards to DISDRODB standards.

    Parameters
    ----------
    sensor_name : str
        Disdrometer sensor name.

    """
    if sensor_name == "OTT_Parsivel":
        d = {
            "Meas_Time": "time",  # TO BE CHECKED
            "Meas_Interval": "sample_interval",
            "RR_Intensity": "rainfall_rate_32bit",
            "RR_Accumulated": "rainfall_accumulated_32bit",
            "RR_Total": "rainfall_amount_absolute_32bit",
            "Synop_WaWa": "weather_code_synop_4680",
            "Synop_WW": "weather_code_synop_4677",
            "Reflectivity": "reflectivity_32bit",
            "Visibility": "mor_visibility",
            "T_Sensor": "sensor_temperature",
            "Sig_Laser": "laser_amplitude",
            "N_Particles": "number_particles",
            "State_Sensor": "sensor_status",
            "E_kin": "rain_kinetic_energy",
            "V_Sensor": "sensor_battery_voltage",
            "I_Heating": "sensor_heating_current",
            "Error_Code": "error_code",
            "Data_Raw": "raw_drop_number",
            "Data_N_Field": "raw_drop_concentration",
            "Data_V_Field": "raw_drop_average_velocity",
        }
    else:
        raise NotImplementedError("DELFT standard implemented only for OTT Parsivel")
    return d


def get_DIVEN_dict(sensor_name: str) -> dict:
    """
    Dictionary mapping from DIVEN standards to DISDRODB standards.

    Parameters
    ----------
    sensor_name : str
        Disdrometer sensor name.

    """
    d = {
        ## Time is to decode
        # 'year'
        # 'month'
        # 'day'
        # 'hour'
        # 'minute'
        # 'second'
        # 'day_of_year'
        "precipitation_flux": "precipitation_rate",
        "solid_precipitation_flux": "snowfall_rate",
        "precipitation_visibility": "mor_visibility",
        "reflectivity": "reflectivity",
        "present_weather_1m": "weather_code_synop_4680",
        "present_weather_5m": "weather_code_synop_4680_5min",
        "max_hail_diameter": "max_hail_diameter",
        "particle_count": "number_particles",
        # 'hydrometeor_type_1m' # Pickering et al., 2019
        # 'hydrometeor_type_5m' # Pickering et al., 2019
        # "measurement_quality": "quality_index",
        # Arrays
        "size_velocity_distribution": "raw_drop_number",
        # "drop_size_distribution": "raw_drop_concentration",
        # "drop_velocity_distribution": "raw_drop_average_velocity",
    }
    return d


def get_ARM_LPM_dict(sensor_name: str) -> dict:
    """
    Dictionary mapping from ARM standards to DISDRODB standards.

    Parameters
    ----------
    sensor_name : str
        Disdrometer sensor name.

    """
    if sensor_name == "Thies_LPM":
        # Dimensions:
        # 'time': 'time # to use
        # "particle_diameter": "diameter bin id ",
        # "particle_fall_velocity": "velocity_bin_id",
        d = {
            # 'base_time': 'base_time',
            # 'time_offset': 'time_offset',
            # 'time_bounds': 'time_bounds',
            # "particle_diameter_bounds",
            # "particle_fall_velocity_bounds"
            "lat": "latitude",
            "lon": "longitude",
            "alt": "altitude",
            "synop_4677_weather_code": "weather_code_synop_4677",
            "metar_4678_weather_code": "weather_code_metar_4678",
            "synop_4680_weather_code": "weather_code_synop_4680",
            "synop_4677_5min_weather_code": "weather_code_synop_4677_5min",
            "metar_4678_5min_weather_code": "weather_code_metar_4678_5min",
            "synop_4680_5min_weather_code": "weather_code_synop_4680_5min",
            "intensity_total_5min": "precipitation_rate_5min",
            "intensity_total": "precipitation_rate",
            "intensity_liquid": "rainfall_rate",
            "intensity_solid": "snowfall_rate",
            "accum_precip": "precipitation_accumulated",
            "maximum_visibility": "mor_visibility",
            "radar_reflectivity": "reflectivity",
            "quality_measurement": "quality_index",
            "max_diameter_hail": "max_hail_diameter",
            "laser_status": "laser_status",
            "static_signal": "static_signal",
            "interior_temperature": "temperature_interior",
            "laser_temperature": "laser_temperature",
            "laser_temperature_analog_status": "laser_temperature_analog_status",
            "laser_temperature_digital_status": "laser_temperature_digital_status",
            "mean_laser_current": "laser_current_average",
            "laser_current_analog_status": "laser_current_analog_status",
            "laser_current_digital_status": "laser_current_digital_status",
            "control_voltage": "control_voltage",
            "optical_control_output": "optical_control_voltage_output",
            "control_output_laser_power_status": "control_output_laser_power_status",
            "voltage_sensor_supply": "sensor_voltage_supply",
            "voltage_sensor_supply_status": "sensor_voltage_supply_status",
            "ambient_temperature": "temperature_ambient",
            "temperature_sensor_status": "temperature_sensor_status",
            "voltage_heating_supply": "current_heating_voltage_supply",
            "voltage_heating_supply_status": "current_heating_voltage_supply_status",
            "pane_heating_laser_head_current": "current_heating_pane_transmitter_head",
            "pane_heating_laser_head_current_status": "current_heating_pane_transmitter_head_status",
            "pane_heating_receiver_head_current": "current_heating_pane_receiver_head",
            "pane_heating_receiver_head_current_status": "current_heating_pane_receiver_head_status",
            "heating_house_current": "current_heating_house",
            "heating_house_current_status": "current_heating_house_status",
            "heating_heads_current": "current_heating_heads",
            "heating_heads_current_status": "current_heating_heads_status",
            "heating_carriers_current": "current_heating_carriers",
            "heating_carriers_current_status": "current_heating_carriers_status",
            "number_particles": "number_particles",
            "number_particles_internal_data": "number_particles_internal_data",
            "number_particles_min_speed": "number_particles_min_speed",
            "number_particles_min_speed_internal_data": "number_particles_min_speed_internal_data",
            "number_particles_max_speed": "number_particles_max_speed",
            "number_particles_max_speed_internal_data": "number_particles_max_speed_internal_data",
            "number_particles_min_diameter": "number_particles_min_diameter",
            "number_particles_min_diameter_internal_data": "number_particles_min_diameter_internal_data",
            "precipitation_spectrum": "raw_drop_number",
            # 'air_temperature',
        }
    elif sensor_name == "OTT_Parsivel2":
        # Coords
        # - time
        # - raw_fall_velocity
        # - particle_size
        d = {
            # 'base_time': 'base_time',
            # 'time_offset'
            # 'class_size_width',
            # 'fall_velocity_calculated',
            "lat": "latitude",
            "lon": "longitude",
            "alt": "altitude",
            "weather_code": "weather_code_synop_4680",
            "equivalent_radar_reflectivity_ott": "reflectivity_32bit",
            "mor_visibility": "mor_visibility",
            "sensor_temperature": "sensor_temperature",
            "laserband_amplitude": "laser_amplitude",
            "heating_current": "sensor_heating_current",
            "sensor_voltage": "sensor_battery_voltage",
            "number_detected_particles": "number_particles",
            "raw_spectrum": "raw_drop_number",
            "number_density_drops": "raw_drop_concentration",
            "snow_depth_intensity": "snowfall_rate",  # Only available > 2019
            # ARM retrievals
            # 'moment1',
            # 'moment2',
            # 'moment3',
            # 'moment4',
            # 'moment5',
            # 'moment6',
            # 'diameter_min':
            # 'diameter_max':
            # 'median_volume_diameter'
            # 'intercept_parameter',
            # 'slope_parameter'
            # 'liquid_water_content',
            # 'liquid_water_distribution_mean'
            # 'precip_rate':
            # 'equivalent_radar_reflectivity',
            # Possible QC variables
            # 'qc_time',
            # 'qc_precip_rate':
            # 'qc_number_detected_particles':
            # 'qc_mor_visibility':
            # 'qc_heating_current':
            # 'qc_snow_depth_intensity':
            # 'qc_laserband_amplitude':
            # 'qc_weather_code':
            # 'qc_equivalent_radar_reflectivity_ott':
            # 'qc_sensor_voltage':
        }
    else:
        raise NotImplementedError
    return d



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

# File content
# - Functions to process raw data files into L0A Apache parquet.

# -----------------------------------------------------------------------------.
import os
import glob
from xmlrpc.server import list_public_methods
import pandas as pd
import dask.dataframe as dd
import logging
import tarfile
from typing import Union

from disdrodb.L0.standards import get_L0A_dtype
from disdrodb.L0.check_standards import check_L0A_standards, check_L0A_column_names
from disdrodb.L0.io import _remove_if_exists
from disdrodb.L0.io import infer_station_id_from_fpath
from disdrodb.utils.logger import log_info, log_warning, log_debug

logger = logging.getLogger(__name__)

# TODO:
# Renaming:
# - read_L0A_raw_file_list --> read_raw_file_list ?
# - write_df_to_parquet --> write_L0A?
# Remove or refactor
# - read_raw_data_zipped


def check_glob_pattern(pattern: str) -> None:
    """Check if the input parameters is a string and if it can be used as pattern.

    Parameters
    ----------
    pattern : str
        String to be checked.

    Raises
    ------
    TypeError
        The input parameter is not a string.
    ValueError
        The input parameter can not be used as pattern.
    """
    if not isinstance(pattern, str):
        raise TypeError("Expect pattern as a string.")
    if pattern[0] == "/":
        raise ValueError("glob_pattern should not start with /")


def _get_file_list(raw_dir: str, glob_pattern) -> list:
    """Get the list of files from a directory based on pattern.

    Parameters
    ----------
    raw_dir : _type_
        Directory of the raw dataset.
    glob_pattern : _type_
        Pattern to match

    Returns
    -------
    list
        List of file paths.
    """
    check_glob_pattern(glob_pattern)
    glob_fpath_pattern = os.path.join(raw_dir, glob_pattern)
    list_fpaths = sorted(glob.glob(glob_fpath_pattern))
    return list_fpaths


def get_file_list(raw_dir, glob_pattern, verbose=False, debugging_mode=False):
    """Get the list of files from a directory based on input parameters.

    Parameters
    ----------
    raw_dir : str
        Directory of the campaign where to search for files.
    glob_pattern : str or list
        Glob pattern to search for files. Can also be a list of glob patterns.
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is False.
    debugging_mode : bool, optional
        If True, it select maximum 3 files for debugging purposes.
        The default is False.

    Returns
    -------
    list_fpaths : list
        List of files file paths.

    """
    if not isinstance(glob_pattern, (str, list)):
        raise ValueError("'glob_pattern' must be a str or list of strings.")
    if isinstance(glob_pattern, str):
        glob_pattern = [glob_pattern]

    # Retrieve filepaths list
    list_fpaths = [_get_file_list(raw_dir, pattern) for pattern in glob_pattern]
    list_fpaths = [x for xs in list_fpaths for x in xs]  # flatten list

    # Check there are files
    n_files = len(list_fpaths)
    if n_files == 0:
        glob_fpath_patterns = [
            os.path.join(raw_dir, pattern) for pattern in glob_pattern
        ]
        raise ValueError(f"No file found at t {glob_fpath_patterns}.")

    # Check there are not directories (or other strange stuffs) in list_fpaths
    # TODO [KIMBO]

    # Subset file_list if debugging_mode
    if debugging_mode:
        max_files = min(3, n_files)
        list_fpaths = list_fpaths[0:max_files]

    # Log
    n_files = len(list_fpaths)
    msg = f" - {n_files} files to process in {raw_dir}"
    if verbose:
        print(msg)
    logger.info(msg)

    # Return file list
    return list_fpaths


####---------------------------------------------------------------------------.
#### Dataframe creation
def preprocess_reader_kwargs(reader_kwargs: dict, lazy: bool = True) -> dict:
    """Define a dictionary with the parameters required for reading the raw data with Pandas or Dask.

    Parameters
    ----------
    reader_kwargs : dict
        Initial parameter dictionary.
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.
    Returns
    -------
    dict
        Parameter dictionary that matches either Pandas or Dask.
    """
    # Remove dtype key
    # - The dtype is enforced to be 'object' in the read function !
    reader_kwargs.pop("dtype", None)

    # Preprocess the reader_kwargs
    reader_kwargs = reader_kwargs.copy()
    if reader_kwargs.get(
        "zipped"
    ):  # TODO: to provide example of this case for testing. I guess could be refactored
        reader_kwargs.pop("zipped", None)
        reader_kwargs.pop("blocksize", None)
        reader_kwargs.pop("file_name_to_read_zipped", None)
    if lazy:
        reader_kwargs.pop("index_col", None)
    if not lazy:
        reader_kwargs.pop("blocksize", None)
        reader_kwargs.pop("assume_missing", None)
    return reader_kwargs


def concatenate_dataframe(
    list_df: list, verbose: bool = False, lazy: bool = True
) -> Union[pd.DataFrame, dd.DataFrame]:
    """Concatenate a list of dataframes.

    Parameters
    ----------
    list_df : list
        List of dataframes.
    verbose : bool, optional
        If True, print messages.
        If False, no print.
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.

    Returns
    -------
    Union[pd.DataFrame,dd.DataFrame]
        Concatenated dataframe.

    Raises
    ------
    ValueError
        Concatenation can not be done.
    """
    # Import dask or pandas
    if lazy:
        import dask.dataframe as dd
    else:
        import pandas as dd
    # Log
    msg = " - Concatenation of dataframes started."
    log_info(logger, msg, verbose)
    # Concatenate the dataframe
    try:
        df = dd.concat(list_df, axis=0, ignore_index=True)
        # Drop duplicated values
        df = df.drop_duplicates(subset="time")
        # Sort by increasing time
        df = df.sort_values(by="time")

    except (AttributeError, TypeError) as e:
        msg = f" - Can not create concat data files. \n Error: {e}"
        logger.error(msg)
        raise ValueError(msg)
    # Log
    msg = " - Concatenation of dataframes has finished."
    log_info(logger, msg, verbose)
    # Return dataframe
    return df


def cast_column_dtypes(df: pd.DataFrame, sensor_name: str) -> pd.DataFrame:
    """Convert 'object' dataframe columns into DISDRODB L0A dtype standards.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    pd.DataFrame
        Dataframe with corrected columns types.
    """

    # Cast dataframe to dtypes
    dtype_dict = get_L0A_dtype(sensor_name)
    # Ensure time column is saved with seconds resolution
    dtype_dict["time"] = "M8[s]"
    # Get dataframe column names
    columns = list(df.columns)
    # Cast dataframe columns
    for column in columns:
        try:
            df[column] = df[column].astype(dtype_dict[column])
        except KeyError:
            # TODO: this must be captured before casting !!!
            # If column dtype is not into L0A_encodings.yml, assign object
            df[column] = df[column].astype("object")
        except ValueError as e:
            raise (f"ValueError: The column {column} has {e}")
    return df


def read_raw_data(
    filepath: str, column_names: list, reader_kwargs: dict, lazy: bool = True
) -> pd.DataFrame:
    """Read raw data into a dataframe

    Parameters
    ----------
    filepath : str
        Raw file path.
    column_names : list
        Column names.
    reader_kwargs : dict
        Dask or Pandas reading parameters
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.

    Returns
    -------
    pandas.DataFrame or dask.DataFrame
        Pandas or dask dataframe
    """
    # Preprocess reader_kwargs
    reader_kwargs = preprocess_reader_kwargs(reader_kwargs, lazy=lazy)

    # Enforce all raw files columns with dtype = 'object'
    dtype = "object"

    # Read with pandas or dask
    # - Dask
    if lazy:
        try:
            df = dd.read_csv(filepath, names=column_names, dtype=dtype, **reader_kwargs)
        except dd.errors.EmptyDataError:
            msg = f" - Is empty, skip file: {filepath}"
            logger.exception(msg)
            print(msg)
            pass
    # Pandas
    else:
        reader_kwargs.pop("blocksize", None)
        try:
            df = pd.read_csv(filepath, names=column_names, dtype=dtype, **reader_kwargs)
        except pd.errors.EmptyDataError:
            msg = f" - Is empty, skip file: {filepath}"
            logger.exception(msg)
            print(msg)
            pass
    # Return dataframe
    return df


def read_raw_data_zipped(
    filepath: str, column_names: list, reader_kwargs: dict, lazy: bool = True
) -> pd.DataFrame:
    """Read zipped raw data into a dataframe.
    Used because some campaign has tar with multiple files inside,
    and in some situation only one files has to be read.
    Tar reading work only with pandas.
    Put the only file name to read into file_name_to_read_zipped variable,
    if file_name_to_read_zipped is none, all the tar contenet will be
    read and concat into a single dataframe.


    Parameters
    ----------
    filepath : str
        Raw file path.
    column_names : list
        Column names.
    reader_kwargs : dict
        Dask or Pandas reading parameters
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.

    Returns
    -------
    pandas.DataFrame or dask.DataFrame
        Pandas or dask dataframe

    Raises
    ------
    pd.errors.EmptyDataError
        File is empty
    pd.errors.ParserError
        File can not be read
    UnicodeDecodeError
        File can not be decoded
    """

    df = pd.DataFrame()
    tar = tarfile.open(filepath)

    file_name_to_read_zipped = reader_kwargs.get("file_name_to_read_zipped")

    # Loop tar files
    for file in tar.getnames():
        # Check if pass only particular file to read
        if file_name_to_read_zipped is not None:
            if file.endswith(file_name_to_read_zipped):
                filepath = file
            else:
                continue

        try:
            # If need only to read one file, exit loop file in tar
            if file_name_to_read_zipped is not None:
                # Read the data
                df = read_raw_data(
                    filepath=tar.extractfile(filepath),
                    column_names=column_names,
                    reader_kwargs=reader_kwargs,
                    lazy=lazy,
                )
                break
            else:
                # Read the data
                df_temp = read_raw_data(
                    filepath=tar.extractfile(file),
                    column_names=column_names,
                    reader_kwargs=reader_kwargs,
                    lazy=lazy,
                )

                # Concat all files in tar
                df = df.append(df_temp)

        except pd.errors.EmptyDataError:
            msg = f" - Is empty, skip file: {file}"
            logger.exception(msg)
            raise pd.errors.EmptyDataError(msg)
            pass
        except pd.errors.ParserError:
            msg = f" - Cannot parse, skip file: {file}"
            logger.exception(msg)
            raise pd.errors.ParserError(msg)
            pass
        except UnicodeDecodeError:
            msg = f" - Unicode error, skip file: {file}"
            logger.exception(msg)
            raise UnicodeDecodeError(msg)
            pass

    # Close zipped file
    tar.close()

    return df


def read_L0A_raw_file_list(
    file_list: Union[list, str],
    column_names: list,
    reader_kwargs: dict,
    sensor_name: str,
    verbose: bool,
    df_sanitizer_fun: object = None,
    lazy: bool = False,
) -> Union[pd.DataFrame, dd.DataFrame]:
    """Read and parse a list for raw files into a dataframe.

    Parameters
    ----------
    file_list : Union[list,str]
        File(s) path(s)
    column_names : list
        Columns names.
    reader_kwargs : dict
        Dask or Pandas reading parameters.
    sensor_name : str
        Name of the sensor.
    verbose : bool
        Wheter to verbose the processing.
    df_sanitizer_fun : object, optional
        Sanitizer function to format the datafame.
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.

    Returns
    -------
    Union[pd.DataFrame,dd.DataFrame]
        Dataframe

    Raises
    ------
    ValueError
        Input parameters can not be used or the raw file can not be processed.

    """

    # ------------------------------------------------------.
    # ### Checks arguments
    if df_sanitizer_fun is not None:
        if not callable(df_sanitizer_fun):
            raise ValueError("'df_sanitizer_fun' must be a function.")
        # TODO check df_sanitizer_fun has only lazy and df arguments !

    if isinstance(file_list, str):
        file_list = [file_list]
    if len(file_list) == 0:
        raise ValueError("'file_list' must contains at least 1 filepath.")

    # ------------------------------------------------------.
    ### - Get station id from file_list
    station_id = infer_station_id_from_fpath(file_list[0])

    # ------------------------------------------------------.
    ### - Loop over all raw files
    n_files = len(file_list)
    processed_file_counter = 0
    list_skipped_files_msg = []
    list_df = []
    for filepath in file_list:
        # Try to process a raw file
        try:
            # -----------------------------------------------------------------.
            # ---------------> THIS SHOULD BE REFACTORED  <-------------------.
            # Open the zip and choose the raw file (for GPM campaign)
            if reader_kwargs.get("zipped"):
                df = read_raw_data_zipped(
                    filepath=filepath,
                    column_names=column_names,
                    reader_kwargs=reader_kwargs,
                    lazy=lazy,
                )

            else:
                # Read the data
                df = read_raw_data(
                    filepath=filepath,
                    column_names=column_names,
                    reader_kwargs=reader_kwargs,
                    lazy=lazy,
                )

            # Check if file empty
            if len(df.index) == 0:
                msg = f" - {filepath} is empty and has been skipped."
                log_warning(logger, msg, verbose)
                list_skipped_files_msg.append(msg)
                continue

            # Check column number, ignore if columns_names empty
            if len(column_names) != 0:
                if len(df.columns) != len(column_names):
                    msg = (
                        f" - {filepath} has wrong columns number, and has been skipped."
                    )
                    log_warning(logger, msg, verbose)
                    list_skipped_files_msg.append(msg)
                    continue
            # -------------------> TILL HERE   <------------------------------.
            # ----------------------------------------------------------------.
            # Sanitize the dataframe with a custom function
            if df_sanitizer_fun is not None:
                df = df_sanitizer_fun(df, lazy=lazy)

            # -------------------------------------------------------.
            # TODO: Make a check to detect that the number of columns does not vary !!!!
            # - Might facilitate identification of buzzy stuffs
            # - Ensure that df_sanitizer_fun capture all problems

            # ------------------------------------------------------.
            # Check column names met DISDRODB standards
            check_L0A_column_names(df, sensor_name=sensor_name)

            # ------------------------------------------------------.
            # Cast dataframe to dtypes
            # - TODO: I would move that after the concatenation !!!
            df = cast_column_dtypes(df, sensor_name=sensor_name)

            # ------------------------------------------------------.
            # Append dataframe to the list
            list_df.append(df)

            # Update the logger
            processed_file_counter += 1
            logger.debug(
                f"{processed_file_counter} / {n_files} processed successfully. File name: {filepath}"
            )

        # If processing of raw file fails
        except (Exception, ValueError) as e:
            # Update the logger
            msg = f" - {filepath} has been skipped. \n -- The error is: {e}."
            log_warning(logger, msg, verbose)
            list_skipped_files_msg.append(msg)

    # Update logger
    msg = f" - {len(list_skipped_files_msg)} of {n_files} have been skipped."
    log_info(logger, msg, verbose)
    logger.info("---")
    logger.info(msg)
    logger.info("---")

    ##----------------------------------------------------------------.
    #### - Concatenate the dataframe
    if len(list_df) == 0:
        raise ValueError(f"No dataframe to return. Impossible to parse {file_list}.")
    if len(list_df) > 1:
        df = concatenate_dataframe(list_df, verbose=verbose, lazy=lazy)
    else:
        df = list_df[0]

    # ------------------------------------------------------.
    #### - Filter out problematic data reported in issue file
    # TODO: [TEST IMPLEMENTATION] remove_problematic_timestamp in dev/TODO_issue_code.py
    # issue_dict = read_issue(raw_dir, station_id)
    # df = remove_problematic_timestamp(df, issue_dict, verbose)

    # ------------------------------------------------------.
    # Return the dataframe
    return df


####---------------------------------------------------------------------------.
#### Parquet Writer
def _write_to_parquet(
    df: Union[pd.DataFrame, dd.DataFrame], fpath: str, force: bool = False
):

    import pandas as pd
    import dask.dataframe

    # -------------------------------------------------------------------------.
    # Check if a file already exists (and remove if force=True)
    _remove_if_exists(fpath, force=force)
    # Cannot create the station folder, so has to be created manually
    from disdrodb.L0.io import _create_directory

    _create_directory(os.path.dirname(fpath))

    # -------------------------------------------------------------------------.
    # Define writing options
    compression = "snappy"  # 'gzip', 'brotli, 'lz4', 'zstd'
    row_group_size = 100000
    engine = "pyarrow"

    # -------------------------------------------------------------------------.
    # Save to parquet
    # - If Pandas df
    if isinstance(df, pd.DataFrame):
        try:
            df.to_parquet(
                fpath,
                engine=engine,
                compression=compression,
                row_group_size=row_group_size,
            )
            logger.info(
                f"The Pandas Dataframe has been written as an Apache Parquet file to {fpath}."
            )
        except (Exception) as e:
            msg = f" - The Pandas DataFrame cannot be written as an Apache Parquet file. The error is: \n {e}."
            logger.exception(msg)
            raise ValueError(msg)

    # - If Dask DataFrame
    elif isinstance(df, dask.dataframe.DataFrame):
        # https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetWriter.html
        try:
            # df.repartition(npartitions=1)
            _ = df.to_parquet(
                fpath,
                schema="infer",
                engine=engine,
                row_group_size=row_group_size,
                compression=compression,
                write_metadata_file=False,
            )
            logger.info(
                f"The Dask Dataframe has been written as an Apache Parquet file to {fpath}."
            )
        except (Exception) as e:
            msg = f" - The Dask DataFrame cannot be written as an Apache Parquet file. The error is: \n {e}."
            logger.exception(msg)
            raise ValueError(msg)
    else:
        raise NotImplementedError("Pandas or Dask DataFrame is required.")

    # -------------------------------------------------------------------------.


def write_df_to_parquet(
    df: Union[pd.DataFrame, dd.DataFrame],
    fpath: str,
    force: bool = False,
    verbose: bool = False,
):
    """Save the dataframe into an Apache Parquet file.

    Parameters
    ----------
    df : Union[pd.DataFrame,dd.DataFrame]
        Input dataframe.
    fpath : str
        Output file path.
    force : bool, optional
        Whether to overwrite existing data.
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories. This is the default.
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is False.

    Raises
    ------
    ValueError
        The input dataframe can not be written as an Apache Parquet file.
    NotImplementedError
        The input dataframe can not be processed.
    """

    # Log
    msg = " - Conversion to Apache Parquet started."
    log_info(logger, msg, verbose)
    # Write to Parquet
    _write_to_parquet(df=df, fpath=fpath, force=force)
    # Log
    msg = " - Conversion to Apache Parquet ended."
    log_info(logger, msg, verbose)
    # -------------------------------------------------------------------------.
    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

# File content
# - Functions to convert L0A Apache parquet files to L0B netCDF files

# -----------------------------------------------------------------------------.
import os
import yaml
import logging
import numpy as np
import pandas as pd
import dask.array as da
import dask.dataframe as dd
import xarray as xr
from typing import Union

from disdrodb.L0.check_standards import (
    check_sensor_name,
    check_L0B_standards,
    # check_array_lengths_consistency,
)
from disdrodb.L0.standards import (
    get_diameter_bin_center,
    get_diameter_bin_lower,
    get_diameter_bin_upper,
    get_diameter_bin_width,
    get_velocity_bin_center,
    get_velocity_bin_lower,
    get_velocity_bin_upper,
    get_velocity_bin_width,
    get_raw_field_nbins,
    get_raw_field_dim_order,
    get_raw_spectrum_ndims,
    get_L0B_encodings_dict,
    get_time_encoding,
)

logger = logging.getLogger(__name__)


# def get_drop_concentration(arr):
#     # TODO
#     logger.info("Computing raw_drop_concentration from raw spectrum.")
#     return arr[:, :, 0]


# def get_drop_average_velocity(arr):
#     # TODO
#     logger.info("Computing raw_drop_average_velocity from raw spectrum.")
#     return arr[:, 0, :]


def check_L0_raw_fields_available(
    df: Union[pd.DataFrame, dd.DataFrame], sensor_name: str
) -> None:
    """Check the presence of the raw spectrum data according to the type of sensor.

    Parameters
    ----------
    df : Union[pd.DataFrame,da.DataFrame]
        Dataframe
    sensor_name : str
        Name of the sensor.

    Raises
    ------
    ValueError
        Error if the raw spectrum field is missing.
    """
    n_bins_dict = get_raw_field_nbins(sensor_name=sensor_name)
    raw_vars = np.array(list(n_bins_dict.keys()))
    missing_vars = raw_vars[np.isin(raw_vars, list(df.columns), invert=True)]
    if len(missing_vars) > 0:
        raise ValueError(f"The following L0 raw fields are missing: {missing_vars}")


def infer_split_str(string: str) -> str:
    """Infer the delimeter inside a string.

    Parameters
    ----------
    string : str
        Input string.

    Returns
    -------
    str
        Inferred delimiter.
    """
    if len(string) > 0:
        valid_delims = [";", ","]  # here we can add others if needed [|, ... ]
        counts = np.array([string.count(delim) for delim in valid_delims])
        idx_delimiter = np.argmax(counts)
        # If don't find the delimiter, set to None
        # --> The array will not be split, and then raise an error later on
        if counts[idx_delimiter] == 0:
            split_str = None
        else:
            split_str = valid_delims[idx_delimiter]
    else:
        split_str = None  # ''.split(None) output []
    return split_str


def format_string_array(string: str, n_values: int) -> np.array:
    """Split a string with multiple numbers separated by a delimiter into an 1D array.

        e.g. : format_string_array("2,44,22,33",4) will return [ 2. 44. 22. 33.]

    Parameters
    ----------
    string : str
        Input string
    n_values : int
        Expected length of the output array.

    Returns
    -------
    np.array
        array of float
    """

    split_str = infer_split_str(string)
    values = np.array(string.split(split_str))

    # -------------------------------------------------------------------------.
    ## Assumptions !!!
    # If empty list --> Assume no precipitation recorded
    if len(values) == 0:
        values = np.zeros(n_values)
        return values

    # If the length of the array is + 1 than the expected, but the last character of
    #  the string is a delimiter --> Drop the last array value
    if len(values) == (n_values + 1):
        if string[-1] == split_str:
            values = np.delete(values, -1)

    # -------------------------------------------------------------------------.
    # If the length is not as expected --> Assume data corruption
    # --> Return an array with nan
    if len(values) != n_values:
        values = np.zeros(n_values) * np.nan
    else:
        # Ensure string type
        values = values.astype("str")
        # Replace '' with 0
        values[values == ""] = "0"
        # Replace "-9.999" with 0
        values = np.char.replace(values, "-9.999", "0")
        # Cast values to float type
        # --> Note: the disk encoding is specified in the L0B_encodings.yml
        values = values.astype(float)
    return values


def reshape_raw_spectrum_to_2D(
    arr: np.array, n_bins_dict: dict, n_timesteps: int
) -> np.array:
    """Reshape the raw spectrum to 2D.

    Parameters
    ----------
    arr : np.array
        Input array.
    n_bins_dict : dict
        Raw field number of bins.
    n_timesteps : int
        Number of timesteps.

    Returns
    -------
    np.array
        Output array.

    Raises
    ------
    ValueError
        Impossible to reshape the raw_spectrum matrix
    """

    try:
        arr = arr.reshape(
            n_timesteps,
            n_bins_dict["raw_drop_concentration"],
            n_bins_dict["raw_drop_average_velocity"],
        )
    except Exception as e:
        msg = f"Impossible to reshape the raw_spectrum matrix. The error is: \n {e}"
        logger.error(msg)
        print(msg)
        raise ValueError(msg)
    return arr


def retrieve_L0B_arrays(
    df: Union[pd.DataFrame, dd.DataFrame],
    sensor_name: str,
    lazy: bool = True,
    verbose: bool = False,
) -> dict:
    """Retrieves the L0B data matrix.

    Parameters
    ----------
    df : Union[pd.DataFrame,dd.DataFrame]
        Input dataframe
    sensor_name : str
        Name of the sensor
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.
        verbose : bool, optional
        Whether to verbose the processing.
        The default is False.

    Returns
    -------
    dict
        Dictionary with data arrays.

    """

    msg = " - Retrieval of L0B data matrix started."
    if verbose:
        print(msg)
    logger.info(msg)
    # ----------------------------------------------------------.
    # Check L0 raw field availability
    # check_L0_raw_fields_available(df, sensor_name)

    # Retrieve raw fields matrix bins dictionary
    n_bins_dict = get_raw_field_nbins(sensor_name=sensor_name)

    # Retrieve dimension order dictionary
    dims_dict = get_raw_field_dim_order(sensor_name)

    # Retrieve dimension of the raw_drop_number field
    n_dim_spectrum = get_raw_spectrum_ndims(sensor_name)

    # Retrieve number of timesteps
    if lazy:
        n_timesteps = df.shape[0].compute()
    else:
        n_timesteps = df.shape[0]

    # Retrieve available arrays
    dict_data = {}
    unavailable_keys = []
    for key, n_bins in n_bins_dict.items():

        # Check key is available in dataframe
        if key not in df.columns:
            unavailable_keys.append(key)
            continue

        # Ensure is a string
        df_series = df[key].astype(str)

        # Get a numpy array for each row and then stack
        if lazy:
            list_arr = df_series.apply(
                format_string_array, n_values=n_bins, meta=(key, "f8")
            )
            arr = da.stack(list_arr, axis=0)
        else:
            list_arr = df_series.apply(format_string_array, n_values=n_bins)
            arr = np.stack(list_arr, axis=0)

        # For key='raw_drop_number', if 2D ... reshape to 2D matrix
        # - This applies i.e for OTT_Parsivels and ThiesLPM
        # - This does not apply to RD80
        if key == "raw_drop_number" and n_dim_spectrum == 2:
            arr = reshape_raw_spectrum_to_2D(arr, n_bins_dict, n_timesteps)

        # Define dictionary to pass to xr.Dataset
        dims_order = ["time"] + dims_dict[key]
        dict_data[key] = (dims_order, arr)

    # -------------------------------------------------------------------------.
    # Retrieve unavailable keys from the raw spectrum field
    # TODO: This should be performed when the xarray object is created !
    # if len(unavailable_keys) > 0:
    #     if "raw_drop_number" not in list(dict_data.keys()):
    #         raise ValueError(
    #             """The raw spectrum is required to compute the unavailables
    #             'raw_drop_concentration' and 'raw_drop_average_velocity' fields."""
    #         )
    #     if "raw_drop_concentration" in unavailable_keys and n_dim_spectrum == 2:
    #         # TODO: can this be computed for RD80 ?
    #         dict_data["raw_drop_concentration"] = get_drop_concentration(
    #             dict_data["raw_drop_number"]
    #         )
    #     if "raw_drop_average_velocity" in unavailable_keys and n_dim_spectrum == 2:
    #         dict_data["raw_drop_average_velocity"] = get_drop_average_velocity(
    #             dict_data["raw_drop_number"]
    #         )
    # -------------------------------------------------------------------------.
    # Log
    msg = " - Retrieval of L0B data matrices finished."
    if verbose:
        print(msg)
    logger.info(msg)

    # Return
    return dict_data


def get_coords(sensor_name: str) -> dict:
    """Retrieve coordinates.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Dictionary with coordinate arrays.
    """

    check_sensor_name(sensor_name=sensor_name)
    coords = {}
    # Retrieve diameter coords
    coords["diameter_bin_center"] = get_diameter_bin_center(sensor_name=sensor_name)
    coords["diameter_bin_lower"] = (
        ["diameter_bin_center"],
        get_diameter_bin_lower(sensor_name=sensor_name),
    )
    coords["diameter_bin_upper"] = (
        ["diameter_bin_center"],
        get_diameter_bin_upper(sensor_name=sensor_name),
    )
    coords["diameter_bin_width"] = (
        ["diameter_bin_center"],
        get_diameter_bin_width(sensor_name=sensor_name),
    )
    # Retrieve velocity coords (if available)
    if get_velocity_bin_center(sensor_name=sensor_name) is not None:
        coords["velocity_bin_center"] = (
            ["velocity_bin_center"],
            get_velocity_bin_center(sensor_name=sensor_name),
        )
        coords["velocity_bin_lower"] = (
            ["velocity_bin_center"],
            get_velocity_bin_lower(sensor_name=sensor_name),
        )
        coords["velocity_bin_upper"] = (
            ["velocity_bin_center"],
            get_velocity_bin_upper(sensor_name=sensor_name),
        )
        coords["velocity_bin_width"] = (
            ["velocity_bin_center"],
            get_velocity_bin_width(sensor_name=sensor_name),
        )
    return coords


def convert_object_variables_to_string(ds: xr.Dataset) -> xr.Dataset:
    """Convert variables with object dtype to string.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.

    Returns
    -------
    xr.Dataset
        Output dataset.
    """
    for var in ds.data_vars:
        if pd.api.types.is_object_dtype(ds[var]):
            ds[var] = ds[var].astype(str)
    return ds


def create_L0B_from_L0A(
    df: Union[pd.DataFrame, dd.DataFrame],
    attrs: dict,
    lazy: bool = True,
    verbose: bool = False,
) -> xr.Dataset:
    """Transform the L0A dataframe to the L0B xr.Dataset.

    Parameters
    ----------
    df : Union[pd.DataFrame,dd.DataFrame]
        DISDRODB L0A dataframe.
    attrs : dict
        Station metadata.
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used.
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is False.

    Returns
    -------
    xr.Dataset
        DISDRODB L0B dataset.

    Raises
    ------
    ValueError
        Error if the DISDRODB L0B xarray dataset can not be created.
    """

    # Retrieve sensor name
    sensor_name = attrs["sensor_name"]
    # -----------------------------------------------------------.
    # Preprocess raw_spectrum, diameter and velocity arrays if available
    if np.any(
        np.isin(
            ["raw_drop_concentration", "raw_drop_average_velocity", "raw_drop_number"],
            df.columns,
        )
    ):
        # Retrieve dictionary of raw data matrices for xarray Dataset
        data_vars = retrieve_L0B_arrays(df, sensor_name, lazy=lazy, verbose=verbose)
    else:
        data_vars = {}
    # -----------------------------------------------------------.
    # Define other disdrometer 'auxiliary' variables varying over time dimension
    valid_core_fields = [
        "raw_drop_concentration",
        "raw_drop_average_velocity",
        "raw_drop_number",
        "time",
        # longitude and latitude too for moving sensors
    ]
    aux_columns = df.columns[np.isin(df.columns, valid_core_fields, invert=True)]
    if lazy:
        aux_data_vars = {
            column: (["time"], df[column].to_dask_array(lengths=True))
            for column in aux_columns
        }
    else:
        aux_data_vars = {
            column: (["time"], df[column].values) for column in aux_columns
        }
    data_vars.update(aux_data_vars)

    # -----------------------------------------------------------.
    # Drop lat/lon array if present (TODO: in future L0 should not contain it)
    if "latitude" in data_vars:
        _ = data_vars.pop("latitude")
    if "longitude" in data_vars:
        _ = data_vars.pop("longitude")
    if "altitude" in data_vars:
        _ = data_vars.pop("altitude")

    # -----------------------------------------------------------.
    # Define coordinates for xarray Dataset
    coords = get_coords(sensor_name=sensor_name)
    coords["time"] = df["time"].values
    coords["crs"] = attrs["crs"]
    if "latitude" in data_vars:
        coords["latitude"] = data_vars["latitude"]
        _ = data_vars.pop("latitude")
    else:
        coords["latitude"] = attrs["latitude"]
    if "longitude" in data_vars:
        coords["longitude"] = data_vars["longitude"]
        _ = data_vars.pop("longitude")
    else:
        coords["longitude"] = attrs["longitude"]
    if "altitude" in data_vars:
        coords["altitude"] = data_vars["altitude"]
        _ = data_vars.pop("altitude")
    else:
        coords["altitude"] = attrs["altitude"]

    # -----------------------------------------------------------
    # Create xarray Dataset
    try:
        ds = xr.Dataset(
            data_vars=data_vars,
            coords=coords,
            attrs=attrs,
        )
    except Exception as e:
        msg = f"Error in the creation of L1 xarray Dataset. The error is: \n {e}"
        logger.error(msg)
        raise ValueError(msg)

    # Ensure variables with dtype object are converted to string
    ds = convert_object_variables_to_string(ds)

    # -----------------------------------------------------------
    # Check L0B standards
    check_L0B_standards(ds)

    # -----------------------------------------------------------
    # TODO: Replace NA flags

    # -----------------------------------------------------------
    return ds


####--------------------------------------------------------------------------.
#### Writers
def sanitize_encodings_dict(encoding_dict: dict, ds: xr.Dataset) -> dict:
    """Ensure chunk size to be smaller than the array shape.

    Parameters
    ----------
    encoding_dict : dict
        Dictionary containing the encoding to write DISDRODB L0B netCDFs.
    ds : xr.Dataset
        Input dataset.

    Returns
    -------
    dict
        Encoding dictionary.
    """
    for var in ds.data_vars:
        shape = ds[var].shape
        chunks = encoding_dict[var]["chunksizes"]
        if chunks is not None:
            chunks = [
                shape[i] if chunks[i] > shape[i] else chunks[i]
                for i in range(len(chunks))
            ]
            encoding_dict[var]["chunksizes"] = chunks
    return encoding_dict


def rechunk_dataset(ds: xr.Dataset, encoding_dict: dict) -> xr.Dataset:
    """Coerce the dataset arrays to have the chunk size specified in the encoding dictionary.

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset
    encoding_dict : dict
        Dictionary containing the encoding to write the xarray dataset as a netCDF.

    Returns
    -------
    xr.Dataset
        Output xarray dataset
    """

    for var in ds.data_vars:
        chunks = encoding_dict[var].pop("chunksizes")
        if chunks is not None:
            ds[var] = ds[var].chunk(chunks)
    return ds


def set_encodings(ds: xr.Dataset, sensor_name: str) -> xr.Dataset:
    """Apply the encodings to the xarray Dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset.
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    xr.Dataset
        Output xarray dataset.
    """
    # Get encoding dictionary
    encoding_dict = get_L0B_encodings_dict(sensor_name)
    encoding_dict = {k: encoding_dict[k] for k in ds.data_vars}

    # Ensure chunksize smaller than the array shape
    encoding_dict = sanitize_encodings_dict(encoding_dict, ds)

    # Rechunk variables for fast writing !
    # - This pop the chunksize argument from the encoding dict !
    ds = rechunk_dataset(ds, encoding_dict)

    # Set time encoding
    ds["time"].encoding.update(get_time_encoding())

    # Set the variable encodings
    for var in ds.data_vars:
        ds[var].encoding.update(encoding_dict[var])

    return ds


def write_L0B(ds: xr.Dataset, fpath: str, sensor_name: str) -> None:
    """Save the xarray dataset into a NetCDF file.

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset.
    fpath : str
        Output file path.
    sensor_name : str
        Name of the sensor.
    """

    # Ensure directory exist
    os.makedirs(os.path.dirname(fpath), exist_ok=True)

    # Set encodings
    ds = set_encodings(ds, sensor_name)

    # Write netcdf
    ds.to_netcdf(fpath, engine="netcdf4")  # , encoding=encoding_dict)


####--------------------------------------------------------------------------.
#### Summary statistics
def create_summary_statistics(
    ds: xr.Dataset, processed_dir: str, station_id: str, sensor_name: str
) -> None:
    """Create L0 summary statistics and save it into the station info YAML file.

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset.
    processed_dir : str
        Output file path
    station_id : str
        Station ID
    sensor_name : str
        Name of the sensor

    """

    ###-----------------------------------------------------------------------.
    # Initialize dictionary
    stats_dict = {}

    # Infer the sampling interval looking at the difference between timesteps
    dt, counts = np.unique(np.diff(ds.time.values), return_counts=True)
    dt_most_frequent = dt[np.argmax(counts)]
    dt_most_frequent = dt_most_frequent.astype("m8[s]")
    inferred_sampling_interval = dt_most_frequent.astype(int)
    stats_dict["inferred_sampling_interval"] = inferred_sampling_interval

    # Number of years, months, days, minutes
    time = ds.time.values
    n_timesteps = len(time)
    n_minutes = inferred_sampling_interval / 60 * n_timesteps
    n_hours = n_minutes / 60
    n_days = n_hours / 24

    stats_dict["n_timesteps"] = n_timesteps
    stats_dict["n_minutes"] = n_minutes
    stats_dict["n_hours"] = n_hours
    stats_dict["n_days"] = n_days

    # Add start_time and end_time
    start_time = pd.DatetimeIndex(time[[0]])
    end_time = pd.DatetimeIndex(time[[-1]])
    years = np.unique([start_time.year, end_time.year])
    if len(years) == 1:
        years_coverage = str(years[0])
    else:
        years_coverage = str(years[0]) + "-" + str(years[-1])

    stats_dict["years_coverage"] = years_coverage
    stats_dict["start_time"] = start_time[0].isoformat()
    stats_dict["end_time"] = end_time[0].isoformat()

    ###-----------------------------------------------------------------------.
    # TODO: Create and save image with temporal coverage
    # --> Colored using quality flag from sensor_status if available ?

    ###-----------------------------------------------------------------------.
    # TODO STATISTICS
    # --> Requiring deriving stats from raw spectrum

    # diameter_min, diameter_max, diameter_sum

    # Total rain events

    # Total rainy minutes

    # Total dry minutes

    # Number of dry/rainy minutes

    ###-----------------------------------------------------------------------.
    # Save to info.yaml
    info_path = os.path.join(processed_dir, "info", station_id + ".yml")
    with open(info_path, "w") as f:
        yaml.dump(stats_dict, f, sort_keys=False)

    return None
